/*****************************************************************************

    pam_locallogin - check if user is logged in locally according to utmp
    Copyright (C) 2004  Egmont Koblinger <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <utmp.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD
#include <security/pam_modules.h>

static int locallogin (pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	const char *user;
	int ok;
	struct utmp utmp;
	if (pam_get_user(pamh, &user, NULL) != PAM_SUCCESS) return -1;

	/*
	   utmpname() and the getut*() functions really suck.
	   Their _r counterparts also suck, they still cannot be safely
	   used without reliying on or modifying some global variables
	   such as the name of the utmp file.
	*/

	int fd = open("/var/run/utmp", O_RDONLY);
	if (fd < 0) return -1;
	while (read(fd, &utmp, sizeof(utmp)) == sizeof(utmp)) {
		if (utmp.ut_type != USER_PROCESS) continue;
		if (strncmp(utmp.ut_user, user, UT_NAMESIZE)) continue;
		ok = 0;
		if (!strncmp(utmp.ut_line, "tty", 3) && (utmp.ut_line[3] >= '1'  && utmp.ut_line[3] <= '9')) ok = 1;
		if (!strncmp(utmp.ut_line, "vc/", 3) && (utmp.ut_line[3] >= '1'  && utmp.ut_line[3] <= '9')) ok = 1;
		if (!strncmp(utmp.ut_line, ":0",  2) && (utmp.ut_line[2] == '\0' || utmp.ut_line[2] == '.')) ok = 1;
		if (!ok) continue;
		close(fd);
		return 0;
	}
	close(fd);
	return -1;
}

PAM_EXTERN
int pam_sm_authenticate (pam_handle_t *pamh, int flags, int argc,
                         const char **argv)
{
	if (locallogin(pamh, flags, argc, argv) < 0) return PAM_AUTH_ERR;
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_setcred (pam_handle_t *pamh, int flags, int argc,
                    const char **argv)
{
	return PAM_IGNORE;
}

PAM_EXTERN
int pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc,
                      const char **argv)
{
	if (locallogin(pamh, flags, argc, argv) < 0) return PAM_AUTH_ERR;
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
                         const char **argv)
{
	if (locallogin(pamh, flags, argc, argv) < 0) return PAM_SESSION_ERR;
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
                          const char **argv)
{
	return PAM_IGNORE;
}

PAM_EXTERN
int pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc,
                      const char **argv)
{
	if (locallogin(pamh, flags, argc, argv) < 0) return PAM_PERM_DENIED;
	return PAM_SUCCESS;
}

#ifdef PAM_STATIC
struct pam_module _pam_locallogin_modstruct = {
	"pam_locallogin",
	pam_sm_authenticate,
	pam_sm_setcred,
	pam_sm_acct_mgmt,
	pam_sm_open_session,
	pam_sm_close_session,
	pam_sm_chauthtok
};
#endif
